from __future__ import annotations

import ast
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from pathlib import Path

    from better_translation.types import MessageID, TranslatedText


FUNCTION_NAMES_TO_EXTRACT_FROM = (
    "gettext",
    "ngettext",
    "lazy_gettext",
    "lazy_ngettext",
    "_",
    "__",
)


@dataclass(slots=True)
class ExtractedMessage:
    id: MessageID = field(init=False)
    default: TranslatedText | None = field(init=False)
    default_plural: TranslatedText | None = field(init=False)
    context: Any | None = field(init=False)
    extra: dict[str, str | int | None] = field(init=False)

    def __init__(
        self,
        message_id: str,
        /,
        default: TranslatedText | None = None,
        default_plural: TranslatedText | None = None,
        context: Any | None = None,
        **extra: Any,
    ) -> None:
        self.id = message_id
        self.default = default
        self.default_plural = default_plural
        self.context = context
        self.extra = extra


def extract_from_code(
    code: str,
    function_names: Sequence[str] = FUNCTION_NAMES_TO_EXTRACT_FROM,
) -> Iterable[ExtractedMessage]:
    """Extract messages from Python code."""
    function_calls_params: list[dict[str, Any]] = []

    class FunctionCallVisitor(ast.NodeVisitor):
        def visit_Call(self, node: ast.Call) -> None:  # noqa: N802
            if (
                isinstance(node.func, ast.Name)
                and node.func.id in function_names
            ):
                args = [
                    ast.literal_eval(arg)
                    if isinstance(arg, ast.Constant)
                    else None
                    for arg in node.args
                ]
                kwargs = {
                    kw.arg: ast.literal_eval(kw.value)
                    if isinstance(kw.value, ast.Constant)
                    else None
                    for kw in node.keywords
                }
                function_calls_params.append({"args": args, "kwargs": kwargs})

            self.generic_visit(node)

    visitor = FunctionCallVisitor()
    tree = ast.parse(code)
    visitor.visit(tree)

    for function_call_params in function_calls_params:
        args = function_call_params["args"]
        kwargs = function_call_params["kwargs"]

        if None in args:
            continue

        call_params = ", ".join(
            (
                *(f"{value!r}" for value in args),
                *(f"{key}={value!r}" for key, value in kwargs.items()),
            ),
        )
        yield eval(f"ExtractedMessage({call_params})")  # noqa: PGH001


def extract_from_file(
    file_path: Path,
    function_names: Sequence[str] = FUNCTION_NAMES_TO_EXTRACT_FROM,
) -> Iterable[ExtractedMessage]:
    """Extract messages from a Python file."""
    yield from extract_from_code(file_path.read_text(), function_names)


def extract_from_dir(
    directory_path: Path,
    function_names: Sequence[str] = FUNCTION_NAMES_TO_EXTRACT_FROM,
) -> Iterable[ExtractedMessage]:
    """Extract messages from Python files in a directory."""
    for path in directory_path.rglob("*.py"):
        yield from extract_from_file(path, function_names)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from better_translation.extractor import extract_from_dir

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from pathlib import Path

    from better_translation.types import Locale, MessageID, TranslatedText

logger = logging.getLogger(__name__)


@dataclass(slots=True)
class StorageMessage:
    id: MessageID
    default: TranslatedText
    default_plural: TranslatedText
    context: str
    has_plural: bool
    translations: Mapping[Locale, StorageTranslation]
    is_dynamic: bool = False


@dataclass(slots=True)
class StorageTranslation:
    singular: TranslatedText
    plural: TranslatedText


class IStorage(ABC):
    @abstractmethod
    def get_message(self, message_id: MessageID) -> StorageMessage | None:
        """Get a message from the storage."""

    @abstractmethod
    async def load(self) -> None:
        """Load translations from the storage to the memory."""

    @abstractmethod
    async def add_messages(self, messages: Sequence[StorageMessage]) -> None:
        """Add new messages to the storage.

        Ignores messages that are already in the storage.
        """

    @abstractmethod
    async def add_messages_from_dirs(self, *dirs: Path) -> list[StorageMessage]:
        """Add new messages to the storage from the given directories."""


@dataclass(slots=True)
class BaseStorage(IStorage, ABC):
    storage: dict[MessageID, StorageMessage] = field(
        default_factory=dict,
        init=False,
    )
    is_loaded: bool = field(default=False, init=False)

    def get_message(self, message_id: MessageID) -> StorageMessage | None:
        if not self.is_loaded:
            logger.warning(
                "Storage is not loaded, message '%s' is not available",
                message_id,
            )

        return self.storage.get(message_id)

    @staticmethod
    def _extract_messages_from_dir(
        directory_path: Path,
    ) -> Iterable[StorageMessage]:
        for message in extract_from_dir(directory_path):
            yield StorageMessage(
                id=message.id,
                default=message.default or message.id,
                default_plural=message.default_plural or "",
                context=str(message.context) if message.context else "",
                has_plural="n" in message.extra,
                translations={},
            )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from better_translation.integrations.django.models import BaseMessage
from better_translation.storage import (
    BaseStorage,
    StorageMessage,
    StorageTranslation,
)

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path


logger = logging.getLogger(__name__)


@dataclass(slots=True, kw_only=True)
class DjangoStorage(BaseStorage):
    message_model: type[BaseMessage]

    async def load(self) -> None:
        """Load translations from the storage to the memory."""
        logger.info("Loading messages from the database...")

        messages = self.message_model.objects.all().prefetch_related(
            "translations",
        )
        async for message in messages:
            translations = message.translations.all()
            if not translations:
                logger.warning(
                    "Message '%s' has no translations",
                    message,
                )
            else:
                logger.debug(
                    "Message '%s' has '%s' translations",
                    message,
                    len(translations),
                )

            self.storage[message.id] = self._get_storage_message(message)

        self.is_loaded = True

        logger.info("Messages loaded successfully")

    async def add_messages(self, messages: Sequence[StorageMessage]) -> None:
        """Add new messages to the storage.

        Ignores messages that are already in the storage.
        """
        if not self.is_loaded:
            await self.load()

        messages_to_create = set[BaseMessage]()

        for message in messages:
            if message.id in self.storage:
                logger.debug(
                    "Message '%s' is already in the storage",
                    message,
                )
                continue

            logger.info("Adding message '%s' to the storage", message)
            self.storage[message.id] = message
            messages_to_create.add(self._get_message_model(message))

        await self.message_model.objects.abulk_create(messages_to_create)

    async def add_messages_from_dirs(
        self,
        *dirs: Path,
        update_messages_usage: bool = False,
    ) -> list[StorageMessage]:
        messages: list[StorageMessage] = []
        for directory_path in dirs:
            messages.extend(self._extract_messages_from_dir(directory_path))

        await self.add_messages(messages)
        if update_messages_usage:
            await self.update_messages_usage(messages)

        return messages

    async def update_messages_usage(
        self,
        messages: Sequence[StorageMessage],
    ) -> None:
        logger.info("Finding unused messages...")

        messages_ids = {message.id for message in messages}
        unused_message_number = await self.message_model.objects.exclude(
            id__in=messages_ids,
        ).aupdate(is_used=False)

        await self.message_model.objects.filter(
            id__in=messages_ids,
        ).aupdate(is_used=True)

        logger.info(
            "Found %s unused messages",
            unused_message_number,
        )

    @staticmethod
    def _get_storage_message(message: BaseMessage) -> StorageMessage:
        return StorageMessage(
            id=message.id,
            default=message.default,
            default_plural=message.default_plural,
            context=message.context,
            has_plural=message.has_plural,
            is_dynamic=message.is_dynamic,
            translations={
                translation.locale: StorageTranslation(
                    singular=translation.singular,
                    plural=translation.singular,
                )
                for translation in message.translations.all()
            },
        )

    def _get_message_model(
        self,
        storage_message: StorageMessage,
    ) -> BaseMessage:
        return self.message_model(
            id=storage_message.id,
            default=storage_message.default,
            default_plural=storage_message.default_plural,
            context=storage_message.context,
            has_plural=storage_message.has_plural,
            is_dynamic=storage_message.is_dynamic,
        )
